import 'dart:math';

import 'package:flutter/material.dart';
import 'ConvertUtils.dart';

enum LengthUnit {
  px, ///logic pixel
  percent, ///percent of the parent size
  ar, ///Aspect Ratio, forbid width to be auto from now on
  vh, ///percent of the screen width
  vw, ///percent of the screen height
  auto, ///no constraint
}

const Map<String, LengthUnit> lengthUnitMap = {
  "auto": LengthUnit.auto,
  "\%": LengthUnit.percent,
  "px": LengthUnit.px,
  "ar": LengthUnit.ar,
  "vh": LengthUnit.vh,
  "vw": LengthUnit.vw,
};

class Length {

  final double value;
  final LengthUnit unit;

  const Length(this.value, {this.unit = LengthUnit.px});

  @override
  String toString() {
    if(unit==LengthUnit.percent) return "\""+value.toString()+"\%"+"\"";
    return "\""+value.toString() + unit.toString().substring(11)+"\"";
  }

  double toPX(BuildContext context,
      {double constraintParentSize = 100, double constraintWidth = 100,
        double screenWidth, double screenHeight}) {

    switch (unit) {
    //for those items with intrinsic size (like image and text)
    //or means no further constraint at this container level
      case LengthUnit.auto:
        return null;
        break;
      case LengthUnit.px:
        return value;
        break;
      case LengthUnit.ar:
        return value * constraintWidth;
        break;
      case LengthUnit.percent:
        if (constraintParentSize == double.infinity) {
          return double.infinity;
        } else {
          return value/100.0 * constraintParentSize;
        }
        break;
      case LengthUnit.vw:
        return value * screenWidth??MediaQuery.of(context).size.width;
        break;
      case LengthUnit.vh:
        return value * screenHeight??MediaQuery.of(context).size.height;
        break;
    }
  }

  Length copyWith({
    double value,
    LengthUnit unit,
  }) {
    return Length(
      value ?? this.value,
      unit: unit ?? this.unit,
    );
  }

  static double newValue(double px, LengthUnit unit, BuildContext context,
      {double constraintParentSize = 200, double constraintWidth = 100,
        double screenWidth, double screenHeight}) {
    switch (unit) {
      case LengthUnit.auto:
        return null;
      case LengthUnit.px:
        return px ?? 100;
      case LengthUnit.ar:
        return (px ?? 0.5) / constraintWidth;
      case LengthUnit.percent:
        return (px ?? 0.5) / constraintParentSize*100.0;
      case LengthUnit.vw:
        return (px ?? 0.5) / screenWidth??MediaQuery.of(context).size.width;
      case LengthUnit.vh:
        return (px ?? 0.2) / screenHeight??MediaQuery.of(context).size.height;
    }

    return null;
  }
}

class DynamicShadow {
  final Color color;
  final double angle;
  final Length offset;
  final Length blurRadius;

  const DynamicShadow({
    this.color=Colors.white,
    this.angle=0,
    this.offset=const Length(0),
    this.blurRadius=const Length(0),
  });

  @override
  String toString() {
    return "{" +
        '''"color": ''' +
        colorToString(color) +
        ''', "angle": ''' +
        angle.toString() +
        ''', "offset": ''' +
        offset.toString() +
        ''', "blurRadius": ''' +
        blurRadius.toString() +
        "}";
  }

  DynamicShadow copyWith({
    Color color,
    double angle,
    Length offset,
    Length blurRadius,
  }) {
    return DynamicShadow(
      color: color ?? this.color,
      angle: angle ?? this.angle,
      offset: offset?? this.offset,
      blurRadius: blurRadius?? this.blurRadius,
    );
  }

  Shadow toShadow(BuildContext context,
      {double constraintParentSize = 200, double constraintWidth = 100}) {
    double offset=this.offset.toPX(context, constraintParentSize: constraintParentSize, constraintWidth: constraintWidth);
    double blurRadius=this.blurRadius.toPX(context, constraintParentSize: constraintParentSize, constraintWidth: constraintWidth);
    return Shadow(
        color: this.color,
        offset: Offset(offset*cos(angle * pi / 180), offset*sin(angle * pi / 180)),
        blurRadius: blurRadius
    );
  }

  BoxShadow toBoxShadow(BuildContext context,
      {double constraintParentSize = 200, double constraintWidth = 100}) {
    double offset=this.offset.toPX(context, constraintParentSize: constraintParentSize, constraintWidth: constraintWidth);
    double blurRadius=this.blurRadius.toPX(context, constraintParentSize: constraintParentSize, constraintWidth: constraintWidth);
    return BoxShadow(
        color: this.color,
        offset: Offset(offset*cos(angle * pi / 180), offset*sin(angle * pi / 180)),
        blurRadius: blurRadius
    );
  }

}