import 'CssUnit.dart';
import 'package:flutter/material.dart';
import 'dart:convert';
import 'CssStyle.dart';

List<dynamic> parseListInMap(Map<String, dynamic> map, String name,
    {Function func, List<dynamic> defaultValue}) {
  return map.containsKey(name) ? parseList(map[name], func: func) : defaultValue;
}

List<dynamic> parseList(List<dynamic> values, {Function func}) {

  if(func!=null) values=values.map((e) => func(e)).toList();

  return values;
}

///parse 1 level map
dynamic parse(String input, {Function func}) {
  dynamic decoded=json.decode(input);

  if(decoded is Map) {
    if(func!=null) decoded=(decoded as Map).map((key,value) => MapEntry(key, func(value)));
  }else{
    if(func!=null) decoded=func(decoded);
  }
  return decoded;
}

Style parseStyle(String style) {
  return Style.fromJson(style);
}

TextStyle parsetextStyleInMap(Map<String, dynamic> map, String name,
    {TextStyle defaultValue}) {
  return map.containsKey(name) ? parseTextStyle(map[name]) : defaultValue;
}

Length parseLengthInMap(Map<String, dynamic> map, String name,
    {Length defaultValue}) {
  return map.containsKey(name) ? parseLength(map[name]) : defaultValue;
}

Gradient parseGradientInMap(Map<String, dynamic> map, String name,
    {Gradient defaultValue}) {
  return map.containsKey(name) ? parseGradient(map[name]) : defaultValue;
}

DecorationImage parseDecorationImageInMap(Map<String, dynamic> map, String name,
    {Gradient defaultValue}) {
  return map.containsKey(name) ? parseDecorationImage(map[name]) : defaultValue;
}

BoxDecoration parseBoxDecorationInMap(Map<String, dynamic> map, String name,
    {BoxDecoration defaultValue}) {
  return map.containsKey(name) ? parseBoxDecoration(map[name]) : defaultValue;
}

bool parseBoolInMap(Map<String, dynamic> map, String name,
    {bool defaultValue = false}) {
  return map.containsKey(name) ? map[name] : defaultValue;
}

int parseIntInMap(Map<String, dynamic> map, String name, {int defaultValue}) {
  return map.containsKey(name) ? map[name] : defaultValue;
}

double parseDoubleInMap(Map<String, dynamic> map, String name,
    {double defaultValue}) {
  return map.containsKey(name) ? map[name] : defaultValue;
}

Alignment parseAlignmentInMap(Map<String, dynamic> map, String name,
    {Alignment defaultValue = Alignment.topLeft}) {
  return map.containsKey(name) ? parseAlignment(map[name]) : Alignment.topLeft;
}

Color parseColorInMap(Map<String, dynamic> map, String name,
    {Color defaultValue = Colors.white}) {
  return map.containsKey(name) ? parseHexColor(map[name]) : defaultValue;
}

TextAlign parseTextAlignInMap(Map<String, dynamic> map, String name,
    {TextAlign defaultValue = TextAlign.left}) {
  return map.containsKey(name) ? parseTextAlign(map[name]) : defaultValue;
}

Axis parseAxisInMap(Map<String, dynamic> map, String name,
    {Axis defaultValue = Axis.vertical}) {
  return map.containsKey(name) ? parseAxis(map[name]) : defaultValue;
}

String parseStringInMap(Map<String, dynamic> map, String name,
    {String defaultValue}) {
  return map.containsKey(name) ? map[name] : defaultValue;
}

Curve parseCurveInMap(Map<String, dynamic> map, String name,
    {Curve defaultValue}) {
  return map.containsKey(name) ? parseCurve(map[name]) : defaultValue;
}

DynamicShadow parseDynamicShadowInMap(Map<String, dynamic> map, String name,
    {DynamicShadow defaultValue}) {
  return map.containsKey(name) ? parseDynamicShadow(map[name]) : defaultValue;
}

ColorFilter parseColorFilter(String colorFilterString) {
  List<String> rst = colorFilterString
      .substring(colorFilterString.indexOf("("), colorFilterString.indexOf(")"))
      .split(",");
  Color color = parseHexColor(rst[0]);
  BlendMode mode = parseBlendMode(rst[1]);
  return ColorFilter.mode(color, mode);
}

List<Length> parseListLength(List<dynamic> values) {
  List<Length> rst = [];
  for (var value in values) {
    rst.add(parseLength(value));
  }
  return rst;
}


Gradient parseGradient(Map<String, dynamic> map) {
  List<Color> colors = parseListInMap(map, "colors", func: parseHexColor).cast<Color>();
  List<double> stops = parseListInMap(map, "stops").cast<double>();
  Alignment alignment1 = parseAlignmentInMap(map, "alignment1");
  Alignment alignment2 = parseAlignmentInMap(map, "alignment2");
  TileMode tileMode = parseTileMode(map["tileMode"]);
  if (map.containsKey("radius")) {
    double radius = parseDoubleInMap(map, "radius");
    return RadialGradient(
        center: alignment1,
        radius: radius,
        colors: colors,
        stops: stops,
        tileMode: tileMode);
  } else if(map.containsKey("startAngle")) {
    double startAngle =parseDoubleInMap(map, "startAngle");
    double endAngle = parseDoubleInMap(map, "endAngle");
    return SweepGradient(
        center: alignment1,
        startAngle: startAngle,
        endAngle: endAngle,
        colors: colors,
        stops: stops,
        tileMode: tileMode);
  }else {
    return LinearGradient(
        begin: alignment1,
        end: alignment2,
        colors: colors,
        stops: stops,
        tileMode: tileMode);
  }

}

BoxDecoration parseBoxDecoration(Map<String, dynamic> map) {
  Color color = parseColorInMap(map, "color", defaultValue: Colors.white);
  Gradient gradient = parseGradientInMap(map, "gradient", defaultValue: null);
  DecorationImage image =
  parseDecorationImageInMap(map, "image", defaultValue: null);
  return BoxDecoration(color: color, gradient: gradient, image: image);
}

DecorationImage parseDecorationImage(Map<String, dynamic> map) {
  String url = parseStringInMap(map, "url");
  Color color = parseHexColor(map["color"]);
  BlendMode mode = parseBlendMode(map["mode"]);
  BoxFit fit = parseBoxFit(map["fit"]);
  Alignment alignment = parseAlignmentInMap(map, "alignment");
  return DecorationImage(
      image: NetworkImage(url),
      fit: fit,
      alignment: alignment,
      colorFilter: ColorFilter.mode(color, mode));
}

DynamicShadow parseDynamicShadow(Map<String, dynamic> map) {
  Color color = parseHexColor(map['color']);
  double angle = map['angle'];
  Length offset = parseLength(map["offset"]);
  Length blurRadius = parseLength(map['blurRadius']);
  return DynamicShadow(
    color: color,
    angle: angle,
    offset: offset,
    blurRadius: blurRadius,
  );
}

TextAlign parseTextAlign(String textAlignString) {
  TextAlign textAlign = TextAlign.left;
  switch (textAlignString) {
    case "TextAlign.left":
      textAlign = TextAlign.left;
      break;
    case "TextAlign.right":
      textAlign = TextAlign.right;
      break;
    case "TextAlign.center":
      textAlign = TextAlign.center;
      break;
    case "TextAlign.justify":
      textAlign = TextAlign.justify;
      break;
    case "TextAlign.start":
      textAlign = TextAlign.start;
      break;
    case "TextAlign.end":
      textAlign = TextAlign.end;
      break;
    default:
      textAlign = TextAlign.left;
  }
  return textAlign;
}

TextOverflow parseTextOverflow(String textOverflowString) {
  TextOverflow textOverflow = TextOverflow.ellipsis;
  switch (textOverflowString) {
    case "ellipsis":
      textOverflow = TextOverflow.ellipsis;
      break;
    case "clip":
      textOverflow = TextOverflow.clip;
      break;
    case "fade":
      textOverflow = TextOverflow.fade;
      break;
    default:
      textOverflow = TextOverflow.fade;
  }
  return textOverflow;
}

TextDirection parseTextDirection(String textDirectionString) {
  TextDirection textDirection = TextDirection.ltr;
  switch (textDirectionString) {
    case 'ltr':
      textDirection = TextDirection.ltr;
      break;
    case 'rtl':
      textDirection = TextDirection.rtl;
      break;
    default:
      textDirection = TextDirection.ltr;
  }
  return textDirection;
}

Color parseHexColor(String hexColorString) {
  if (hexColorString == null) {
    return null;
  }
  int colorInt = int.parse(hexColorString, radix: 16);
  return Color(colorInt);
}

TextStyle parseTextStyle(Map<String, dynamic> map) {
  //TODO: more properties need to be implemented, such as decoration, decorationColor, decorationStyle, wordSpacing and so on.
  Color color = parseHexColor(map['color']);
  Color backgroundColor= parseHexColor(map["backgroundColor"]);
  String fontFamily = map['fontFamily'];
  //double fontSize = map['fontSize'];
  FontStyle fontStyle = 'FontStyle.italic' == map['fontStyle']
      ? FontStyle.italic
      : FontStyle.normal;
  FontWeight fontWeight = 'FontWeight.w700' == map['fontWeight']
      ? FontWeight.bold
      : FontWeight.normal;
  TextDecoration decoration = parseTextDecoration(map['decoration']);
  Color decorationColor= parseHexColor(map["decorationColor"]);
  TextDecorationStyle decorationStyle = parseTextDecorationStyle(map['decorationStyle']);
  double decorationThickness = map['decorationThickness'];
  double height=map['height'];
  double letterSpacing=map['letterSpacing'];

  return TextStyle(
    height: height,
    letterSpacing: letterSpacing,
    color: color,
    backgroundColor: backgroundColor,
    //fontSize: fontSize,
    fontFamily: fontFamily,
    fontStyle: fontStyle,
    fontWeight: fontWeight,
    decoration: decoration,
    decorationColor: decorationColor,
    decorationStyle: decorationStyle,
    decorationThickness: decorationThickness,
  );
}

Alignment parseAlignment(String alignmentString) {
  Alignment alignment = Alignment.topLeft;
  switch (alignmentString) {
    case 'topLeft':
      alignment = Alignment.topLeft;
      break;
    case 'topCenter':
      alignment = Alignment.topCenter;
      break;
    case 'topRight':
      alignment = Alignment.topRight;
      break;
    case 'centerLeft':
      alignment = Alignment.centerLeft;
      break;
    case 'center':
      alignment = Alignment.center;
      break;
    case 'centerRight':
      alignment = Alignment.centerRight;
      break;
    case 'bottomLeft':
      alignment = Alignment.bottomLeft;
      break;
    case 'bottomCenter':
      alignment = Alignment.bottomCenter;
      break;
    case 'bottomRight':
      alignment = Alignment.bottomRight;
      break;

    default:
      {
        List<String> aligns = alignmentString
            .substring(
            alignmentString.indexOf("("), alignmentString.indexOf(")"))
            .split(",");
        alignment = Alignment(double.parse(aligns[0]), double.parse(aligns[1]));
      }
      break;
  }
  return alignment;
}

const double infinity = 9999999999;

BoxConstraints parseBoxConstraints(Map<String, dynamic> map) {
  double minWidth = 0.0;
  double maxWidth = double.infinity;
  double minHeight = 0.0;
  double maxHeight = double.infinity;

  if (map != null) {
    if (map.containsKey('minWidth')) {
      var minWidthValue = map['minWidth'];

      if (minWidthValue != null) {
        if (minWidthValue >= infinity) {
          minWidth = double.infinity;
        } else {
          minWidth = minWidthValue;
        }
      }
    }

    if (map.containsKey('maxWidth')) {
      var maxWidthValue = map['maxWidth'];

      if (maxWidthValue != null) {
        if (maxWidthValue >= infinity) {
          maxWidth = double.infinity;
        } else {
          maxWidth = maxWidthValue;
        }
      }
    }

    if (map.containsKey('minHeight')) {
      var minHeightValue = map['minHeight'];

      if (minHeightValue != null) {
        if (minHeightValue >= infinity) {
          minHeight = double.infinity;
        } else {
          minHeight = minHeightValue;
        }
      }
    }

    if (map.containsKey('maxHeight')) {
      var maxHeightValue = map['maxHeight'];

      if (maxHeightValue != null) {
        if (maxHeightValue >= infinity) {
          maxHeight = double.infinity;
        } else {
          maxHeight = maxHeightValue;
        }
      }
    }
  }

  return BoxConstraints(
    minWidth: minWidth,
    maxWidth: maxWidth,
    minHeight: minHeight,
    maxHeight: maxHeight,
  );
}

EdgeInsetsGeometry parseEdgeInsetsGeometry(String edgeInsetsGeometryString) {
  //left,top,right,bottom
  if (edgeInsetsGeometryString == null ||
      edgeInsetsGeometryString.trim() == '') {
    return null;
  }
  var values = edgeInsetsGeometryString.split(",");
  return EdgeInsets.only(
      left: double.parse(values[0]),
      top: double.parse(values[1]),
      right: double.parse(values[2]),
      bottom: double.parse(values[3]));
}

CrossAxisAlignment parseCrossAxisAlignment(String crossAxisAlignmentString) {

  crossAxisAlignmentString=crossAxisAlignmentString.substring(18);

  switch (crossAxisAlignmentString) {
    case 'start':
      return CrossAxisAlignment.start;
    case 'end':
      return CrossAxisAlignment.end;
    case 'center':
      return CrossAxisAlignment.center;
    case 'stretch':
      return CrossAxisAlignment.stretch;
    case 'baseline':
      return CrossAxisAlignment.baseline;
  }
  return CrossAxisAlignment.center;
}

MainAxisAlignment parseMainAxisAlignment(String mainAxisAlignmentString) {

  mainAxisAlignmentString=mainAxisAlignmentString.substring(17);

  switch (mainAxisAlignmentString) {
    case 'start':
      return MainAxisAlignment.start;
    case 'end':
      return MainAxisAlignment.end;
    case 'center':
      return MainAxisAlignment.center;
    case 'spaceBetween':
      return MainAxisAlignment.spaceBetween;
    case 'spaceAround':
      return MainAxisAlignment.spaceAround;
    case 'spaceEvenly':
      return MainAxisAlignment.spaceEvenly;
  }
  return MainAxisAlignment.start;
}

WrapAlignment parseWrapAlignment(String wrapAlignmentString) {

  wrapAlignmentString=wrapAlignmentString.substring(14);

  switch (wrapAlignmentString) {
    case 'start':
      return WrapAlignment.start;
    case 'end':
      return WrapAlignment.end;
    case 'center':
      return WrapAlignment.center;
    case 'spaceBetween':
      return WrapAlignment.spaceBetween;
    case 'spaceAround':
      return WrapAlignment.spaceAround;
    case 'spaceEvenly':
      return WrapAlignment.spaceEvenly;
  }
  return WrapAlignment.start;
}

WrapCrossAlignment parseWrapCrossAlignment(String wrapCrossAlignmentString) {

  wrapCrossAlignmentString=wrapCrossAlignmentString.substring(19);

  switch (wrapCrossAlignmentString) {
    case 'start':
      return WrapCrossAlignment.start;
    case 'end':
      return WrapCrossAlignment.end;
    case 'center':
      return WrapCrossAlignment.center;
  }
  return WrapCrossAlignment.start;
}

MainAxisSize parseMainAxisSize(String mainAxisSizeString) =>
    mainAxisSizeString == 'min' ? MainAxisSize.min : MainAxisSize.max;

TextBaseline parseTextBaseline(String parseTextBaselineString) =>
    'alphabetic' == parseTextBaselineString
        ? TextBaseline.alphabetic
        : TextBaseline.ideographic;

VerticalDirection parseVerticalDirection(String verticalDirectionString) =>
    'up' == verticalDirectionString
        ? VerticalDirection.up
        : VerticalDirection.down;

BlendMode parseBlendMode(String blendModeString) {
  if (blendModeString == null || blendModeString.trim().length == 0) {
    return null;
  }

  blendModeString = blendModeString.substring(11);

  switch (blendModeString.trim()) {
    case 'clear':
      return BlendMode.clear;
    case 'src':
      return BlendMode.src;
    case 'dst':
      return BlendMode.dst;
    case 'srcOver':
      return BlendMode.srcOver;
    case 'dstOver':
      return BlendMode.dstOver;
    case 'srcIn':
      return BlendMode.srcIn;
    case 'dstIn':
      return BlendMode.dstIn;
    case 'srcOut':
      return BlendMode.srcOut;
    case 'dstOut':
      return BlendMode.dstOut;
    case 'srcATop':
      return BlendMode.srcATop;
    case 'dstATop':
      return BlendMode.dstATop;
    case 'xor':
      return BlendMode.xor;
    case 'plus':
      return BlendMode.plus;
    case 'modulate':
      return BlendMode.modulate;
    case 'screen':
      return BlendMode.screen;
    case 'overlay':
      return BlendMode.overlay;
    case 'darken':
      return BlendMode.darken;
    case 'lighten':
      return BlendMode.lighten;
    case 'colorDodge':
      return BlendMode.colorDodge;
    case 'colorBurn':
      return BlendMode.colorBurn;
    case 'hardLight':
      return BlendMode.hardLight;
    case 'softLight':
      return BlendMode.softLight;
    case 'difference':
      return BlendMode.difference;
    case 'exclusion':
      return BlendMode.exclusion;
    case 'multiply':
      return BlendMode.multiply;
    case 'hue':
      return BlendMode.hue;
    case 'saturation':
      return BlendMode.saturation;
    case 'color':
      return BlendMode.color;
    case 'luminosity':
      return BlendMode.luminosity;

    default:
      return BlendMode.srcIn;
  }
}

BoxFit parseBoxFit(String boxFitString) {
  if (boxFitString == null) {
    return null;
  }

  boxFitString = boxFitString.substring(7);

  switch (boxFitString) {
    case 'fill':
      return BoxFit.fill;
    case 'contain':
      return BoxFit.contain;
    case 'cover':
      return BoxFit.cover;
    case 'fitWidth':
      return BoxFit.fitWidth;
    case 'fitHeight':
      return BoxFit.fitHeight;
    case 'none':
      return BoxFit.none;
    case 'scaleDown':
      return BoxFit.scaleDown;
  }

  return null;
}

ImageRepeat parseImageRepeat(String imageRepeatString) {
  if (imageRepeatString == null) {
    return null;
  }

  switch (imageRepeatString) {
    case 'repeat':
      return ImageRepeat.repeat;
    case 'repeatX':
      return ImageRepeat.repeatX;
    case 'repeatY':
      return ImageRepeat.repeatY;
    case 'noRepeat':
      return ImageRepeat.noRepeat;

    default:
      return ImageRepeat.noRepeat;
  }
}

Rect parseRect(String fromLTRBString) {
  var strings = fromLTRBString.split(',');
  return Rect.fromLTRB(double.parse(strings[0]), double.parse(strings[1]),
      double.parse(strings[2]), double.parse(strings[3]));
}

FilterQuality parseFilterQuality(String filterQualityString) {
  if (filterQualityString == null) {
    return null;
  }
  switch (filterQualityString) {
    case 'none':
      return FilterQuality.none;
    case 'low':
      return FilterQuality.low;
    case 'medium':
      return FilterQuality.medium;
    case 'high':
      return FilterQuality.high;
    default:
      return FilterQuality.low;
  }
}

StackFit parseStackFit(String value) {
  if (value == null) return null;

  switch (value) {
    case 'loose':
      return StackFit.loose;
    case 'expand':
      return StackFit.expand;
    case 'passthrough':
      return StackFit.passthrough;
    default:
      return StackFit.loose;
  }
}

Overflow parseOverflow(String value) {
  if (value == null) {
    return null;
  }

  switch (value) {
    case 'visible':
      return Overflow.visible;
    case 'clip':
      return Overflow.clip;
    default:
      return Overflow.clip;
  }
}

Axis parseAxis(String axisString) {
  if (axisString == null) {
    return Axis.vertical;
  }

  switch (axisString) {
    case "Axis.horizontal":
      return Axis.horizontal;
    case "Axis.vertical":
      return Axis.vertical;
  }
  return Axis.vertical;
}

TextDecoration parseTextDecoration(String decoration) {
  if (decoration == null) {
    return TextDecoration.none;
  }

  switch (decoration) {
    case "TextDecoration.none":
      return TextDecoration.none;
    case "TextDecoration.underline":
      return TextDecoration.underline;
    case "TextDecoration.lineThrough":
      return TextDecoration.lineThrough;
    case "TextDecoration.overline":
      return TextDecoration.overline;
  }
  return TextDecoration.none;

}

TextDecorationStyle parseTextDecorationStyle(String style) {
  if (style == null) {
    return TextDecorationStyle.solid;
  }

  switch (style) {
    case "TextDecorationStyle.solid":
      return TextDecorationStyle.solid;
    case "TextDecorationStyle.double":
      return TextDecorationStyle.double;
    case "TextDecorationStyle.dotted":
      return TextDecorationStyle.dotted;
    case "TextDecorationStyle.dashed":
      return TextDecorationStyle.dashed;
    case "TextDecorationStyle.wavy":
      return TextDecorationStyle.wavy;
  }
  return TextDecorationStyle.solid;

}

TileMode parseTileMode(String tileModeString) {
  if(tileModeString==null) {
    return TileMode.clamp;
  }

  switch (tileModeString) {
    case "TileMode.clamp":
      return TileMode.clamp;
    case "TileMode.mirror":
      return TileMode.mirror;
    case "TileMode.repeated":
      return TileMode.repeated;
  }

  return TileMode.clamp;

}

Clip parseClipBehavior(String clipBehaviorString) {
  if (clipBehaviorString == null) {
    return Clip.antiAlias;
  }
  switch (clipBehaviorString) {
    case "antiAlias":
      return Clip.antiAlias;
    case "none":
      return Clip.none;
    case "hardEdge":
      return Clip.hardEdge;
    case "antiAliasWithSaveLayer":
      return Clip.antiAliasWithSaveLayer;
  }
  return Clip.antiAlias;
}

Curve parseCurve(String curve) {
  if (curve == null || curve.trim().length == 0) {
    return null;
  }

  //curve = curve.substring(7);

  switch (curve) {
    case "linear":
      return Curves.linear;
    case "bounceIn":
      return Curves.bounceIn;
    case "bounceOut":
      return Curves.bounceOut;
    case "decelerate":
      return Curves.decelerate;
    case "ease":
      return Curves.ease;
    case "easeIn":
      return Curves.easeIn;
    case "easeOut":
      return Curves.easeOut;
    case "elasticIn":
      return Curves.elasticIn;
    case "elasticOut":
      return Curves.elasticOut;
  }

  return Curves.linear;
}

LengthUnit parseLengthUnit(String unit) {
  if (unit == null || unit.trim().length == 0) {
    return null;
  }

  return lengthUnitMap[unit];
}

Length parseLength(String string) {
  if(string==null || string.trim().length == 0) return null;

  final RegExp regExp = RegExp('[a-zA-Z\%]');
  int index=string.indexOf(regExp);

  return Length(double.tryParse(string.substring(0,index)), unit: parseLengthUnit(string.substring(index)));
}
